import logging

from flask import Flask
from guniflask.config import Settings

log = logging.getLogger(__name__)


def make_settings(app: Flask, settings: Settings):
    pass


def init_app(app: Flask, settings: Settings):
    pass



from guniflask.utils.string import string_camelcase, string_lowercase_hyphen, \
    string_lowercase_underscore, string_uppercase_underscore


def test_string_camelcase():
    assert string_camelcase('get_http_response') == 'GetHttpResponse'
    assert string_camelcase('meta-data') == 'MetaData'
    assert string_camelcase('N2O') == 'N2O'


def test_string_lowercase_hyphen():
    assert string_lowercase_hyphen('getHTTPResponse') == 'get-http-response'
    assert string_lowercase_hyphen('MetaData') == 'meta-data'


def test_string_lowercase_underscore():
    assert string_lowercase_underscore('getHTTPResponse') == 'get_http_response'
    assert string_lowercase_underscore('MetaData') == 'meta_data'


def test_string_uppercase_underscore():
    assert string_uppercase_underscore('getHTTPResponse') == 'GET_HTTP_RESPONSE'
    assert string_uppercase_underscore('MetaData') == 'META_DATA'



import os
from os.path import dirname, curdir, join, exists, abspath

from guniflask.config.env import load_app_env


def set_test_env():
    home = os.environ.get('GUNIFLASK_HOME')
    if home is None:
        home = abspath(curdir)
        while True:
            if not exists(join(home, '__init__.py')):
                break
            home = dirname(home)
        os.environ['GUNIFLASK_HOME'] = home
    os.environ.setdefault('GUNIFLASK_DEBUG', '1')
    os.environ.setdefault('GUNIFLASK_ACTIVE_PROFILES', 'dev')
    load_app_env()



from abc import ABCMeta, abstractmethod
from typing import Type, Optional

from guniflask.security.authentication import Authentication


class AuthenticationProvider(metaclass=ABCMeta):
    @abstractmethod
    def authenticate(self, authentication: Authentication) -> Optional[Authentication]:
        pass  # pragma: no cover

    @abstractmethod
    def supports(self, authentication_cls: Type[Authentication]) -> bool:
        pass



from .access_annotation import has_any_authority
from .access_annotation import has_any_role
from .access_annotation import has_authority
from .access_annotation import has_role
from .access_annotation import login_required
from .authentication import Authentication
from .authentication_manager import AuthenticationManager
from .authentication_provider import AuthenticationProvider
from .authentication_token import UserAuthentication
from .context import SecurityContext
from .context import current_user
from .errors import AuthenticationError
from .errors import BadCredentialsError
from .errors import InsufficientAuthenticationError
from .errors import ProviderNotFoundError
from .errors import UsernameNotFoundError
from .jwt import JwtHelper
from .jwt_manager import JwtManager
from .password_encoder import PasswordEncoder
from .preauth_token import BearerTokenExtractor
from .preauth_token import PreAuthenticatedToken
from .provider_manager import ProviderManager
from .user import User
from .user_details import UserDetails
from .web_authentication_details import WebAuthenticationDetails



import socket


def get_local_ip_address():
    s = None
    try:
        s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        s.connect(('224.0.0.0', 80))
        local_ip = s.getsockname()[0]
        return local_ip
    finally:
        if s:
            s.close()



from guniflask.annotation import Annotation, AnnotationUtils


class MethodDefFilter(Annotation):
    def add_filter(self, name, args=None):
        if 'values' not in self:
            self['values'] = []
        self['values'].append({'name': name, 'args': args})

    def get_filters(self):
        if 'values' not in self:
            return []
        return self['values']


def _add_method_def_filter_annotation(func, name, args=None):
    a = AnnotationUtils.get_annotation(func, MethodDefFilter)
    if a is None:
        a = MethodDefFilter()
        AnnotationUtils.add_annotation(func, a)
    a.add_filter(name, args=args)


def before_request(func):
    _add_method_def_filter_annotation(func, 'before_request')
    return func


def after_request(func):
    _add_method_def_filter_annotation(func, 'after_request')
    return func


def app_before_request(func):
    _add_method_def_filter_annotation(func, 'app_before_request')
    return func


def app_after_request(func):
    _add_method_def_filter_annotation(func, 'app_after_request')
    return func


def error_handler(code_or_exception):
    def wrap_func(func):
        _add_method_def_filter_annotation(func, 'errorhandler', args=[code_or_exception])
        return func

    return wrap_func


def app_error_handler(code_or_exception):
    def wrap_func(func):
        _add_method_def_filter_annotation(func, 'app_errorhandler', args=[code_or_exception])
        return func

    return wrap_func



import datetime as dt
import uuid

from guniflask.security import Authentication
from guniflask.security.authentication_manager import AuthenticationManager
from guniflask.security.authentication_token import UserAuthentication
from guniflask.security.errors import InvalidTokenError
from guniflask.security.jwt import JwtHelper
from guniflask.security.user import User


class JwtManager(AuthenticationManager):
    EXP = 'exp'
    JTI = 'jti'
    AUTHORITIES = 'authorities'
    USERNAME = 'username'
    USERINFO = 'userinfo'

    def __init__(self, secret='', public_key=None, private_key=None, algorithm=None,
                 access_token_expires_in=12 * 60 * 60, refresh_token_expires_in=30 * 24 * 60 * 60):
        self.secret = secret
        self.public_key = public_key
        self.private_key = private_key
        self.algorithm = algorithm
        self.access_token_expires_in = access_token_expires_in
        self.refresh_token_expires_in = refresh_token_expires_in

    def create_access_token(self, authorities=None, username=None, **userinfo) -> str:
        expires_in = self.access_token_expires_in
        payload = {
            self.JTI: uuid.uuid4().hex
        }
        exp = dt.datetime.now() + dt.timedelta(seconds=expires_in)
        payload[self.EXP] = exp

        payload[self.AUTHORITIES] = authorities
        payload[self.USERNAME] = username
        payload[self.USERINFO] = userinfo

        return JwtHelper.encode_jwt(
            payload,
            self.secret or self.private_key,
            algorithm=self.algorithm,
        )

    def authenticate(self, authentication: Authentication):
        access_token_value = authentication.principal
        payload = self._decode(access_token_value)
        authorities = payload.get(self.AUTHORITIES)
        username = payload.get(self.USERNAME)
        userinfo = payload.get(self.USERINFO)
        user = User(username=username, authorities=authorities)
        if userinfo and isinstance(userinfo, dict):
            for k, v in userinfo.items():
                setattr(user, k, v)
        user_auth = UserAuthentication(user, authorities=authorities)
        user_auth.authenticate(True)
        return user_auth

    def _decode(self, access_token_value):
        try:
            payload = JwtHelper.decode_jwt(
                access_token_value,
                self.secret or self.public_key,
                algorithms=[self.algorithm] if self.algorithm else None,
            )
        except Exception as e:
            raise InvalidTokenError(e)
        return payload



from flask import abort
from guniflask.context import service

from ..config.jwt_config import jwt_manager


@service
class AccountService:
    accounts = {
        'root': {
            'authorities': ['role_admin'],
            'password': '123456',
        }
    }

    def login(self, username: str, password: str):
        if username not in self.accounts or self.accounts[username]['password'] != password:
            return abort(403)
        account = self.accounts[username]
        token = jwt_manager.create_access_token(authorities=account['authorities'], username=username)
        return {
            'username': username,
            'access_token': token,
        }

    def get(self, username: str):
        if username not in self.accounts:
            return abort(404)
        return {
            'username': username,
            'authorities': self.accounts[username]['authorities']
        }



from typing import List

from guniflask.beans.lifecycle import InitializingBean
from guniflask.security.authentication import Authentication
from guniflask.security.authentication_manager import AuthenticationManager
from guniflask.security.authentication_provider import AuthenticationProvider
from guniflask.security.errors import AuthenticationError, ProviderNotFoundError


class ProviderManager(AuthenticationManager, InitializingBean):
    def __init__(self, parent: AuthenticationManager = None, providers: List[AuthenticationProvider] = None):
        self._parent = parent
        self._providers: List[AuthenticationProvider] = []
        if providers:
            for p in providers:
                self._providers.append(p)

    def authenticate(self, authentication: Authentication):
        to_test = type(authentication)
        last_error = None
        result = None

        for provider in self._providers:
            if not provider.supports(to_test):
                continue
            try:
                result = provider.authenticate(authentication)
            except AuthenticationError as e:
                last_error = e
            else:
                if result is not None:
                    if result.details is None:
                        result.details = authentication.details
                    break

        if result is None and self._parent is not None:
            try:
                result = self._parent.authenticate(authentication)
            except ProviderNotFoundError:
                pass
            except AuthenticationError as e:
                last_error = e

        if result is None:
            if last_error is None:
                last_error = ProviderNotFoundError(f'No authentication provider for {to_test.__name__}')
            raise last_error

        return result

    def after_properties_set(self):
        self._check_state()

    def _check_state(self):
        if not self._parent and not self._providers:
            raise ValueError('A parent authentication manager or a list of authentication provides is required')
